#include <cyg/hal/hal_cache.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/kernel/kapi.h>
#include <cyg/error/codes.h>
#include <cyg/io/io.h>
#include <cyg/io/serialio.h>
#include <cyg/io/file.h>
#include <net/etherio.h>
#include <stdlib.h>
#include <network.h>
#include <stdio.h>
#include <errno.h>

#include "../dm.h"
#include "../ident.h"

extern int select(int, fd_set *, fd_set *, fd_set *, struct timeval *tv);
extern int cyg_select_with_abort(int, fd_set *, fd_set *, fd_set *, struct timeval *tv);
extern void cyg_select_abort(void);
extern int close(int);

//============================================================================
// A sample extenstion to DM SocketServer
//============================================================================

// Notice that this file exposes no symbol names.  Everything is static.

// product ID info (we don't use it in this module, but you might want to
extern tIdent ident;

// state for our extension threads -- we're only going to have
// two, but you can have more if you like.

static cyg_thread_entry_t echoTask;
static unsigned char echoStack[4096];
static cyg_handle_t  echoHandle;
static cyg_thread    echoThread;

static cyg_thread_entry_t bgTask;
static unsigned char bgStack[4096];
static cyg_handle_t  bgHandle;
static cyg_thread    bgThread;

// global flag set by TCP/IP stack init routine, in case you
// want to know whether TCP/IP network is up or not

extern int TcpStackOk;

// declare the functions that run as tasks

// diag printf routine that prints messages to internal UART
extern void diag_printf(const char *fmt, ...);

// initialization routine called before scheduler is running
void extensionInit0(void)
{
  diag_printf("ExtensionInit0()\n");
  cyg_thread_create(8, echoTask, (cyg_addrword_t)0, "Extension - Echo",
                    (void *)echoStack, 
                    sizeof echoStack, 
                    &echoHandle,
                    &echoThread);
  cyg_thread_resume(echoHandle);
  
  cyg_thread_create(8, bgTask, (cyg_addrword_t)0, "Extension - Bg",
                    (void *)bgStack, 
                    sizeof bgStack, 
                    &bgHandle,
                    &bgThread);
  cyg_thread_resume(bgHandle);
}

// initialization routine called after scheduler is running
// and network initialization has completed (either with
// success or failure)
void extensionInit1(void)
{
  diag_printf("extensionInit1()\n");
  diag_printf("TCP stack OK = %d\n",TcpStackOk);
}

// someplace to go when we decide to give up
static void done(void)
{
  diag_printf("extension done!\n");
  while (1)
    cyg_thread_delay(100);
}

// count of bytes echoed

static unsigned totalBytes;

// this is a simple thread that echos data on a serial port

static void echoTask(cyg_addrword_t data)
{
  unsigned char buf[512];
  cyg_uint32 len;
  Cyg_ErrNo status;
  cyg_io_handle_t ioHandle;
  cyg_serial_info_t serConfig;
  
  diag_printf("echoTask Beginning execution\n");

  // acquire ownership of port
  
  cyg_semaphore_wait(&devMastPort[0].ownership);
  
  // The socket server code may be currently blocked in
  // select(), and to be polite, we should wake it up so that
  // it can shut down listening sockets if it wants to.
  
  cyg_select_abort();
  

  diag_printf("echoTask acquired port 0\n");

  // configure port
  
  ioHandle = devMastPort[0].handle;
  
  len = sizeof serConfig;
  status = cyg_io_get_config(ioHandle, CYG_IO_GET_CONFIG_SERIAL_INFO, &serConfig, &len);
  
  if (status != ENOERR)
    {
      diag_printf("ERROR, cyg_io_get_config returned %d: %s\n",status,strerror(status));
      done();
    }

  serConfig.flags &= ~CYG_SERIAL_FLAGS_RTSCTS;
  serConfig.flags &= ~CYG_SERIAL_FLAGS_XONXOFF_RXFLOW;
  serConfig.flags &= ~CYG_SERIAL_FLAGS_XONXOFF_TXFLOW;

  serConfig.flags = CYG_SERIAL_FLAGS_RD_SEMIBLOCK;
  
  serConfig.baud = 9600;
  serConfig.stop = CYGNUM_SERIAL_STOP_1;
  serConfig.parity = CYGNUM_SERIAL_PARITY_NONE;
  serConfig.word_length = CYGNUM_SERIAL_WORD_LENGTH_8;
  
  len = sizeof serConfig;
  status = cyg_io_set_config(ioHandle, CYG_IO_SET_CONFIG_SERIAL_INFO, &serConfig, &len);
  
  if (status != ENOERR)
    {
      diag_printf("ERROR, cyg_io_get_config returned %d: %s\n",status,strerror(status));
      done();
    }

  // set port interface mode to RS-232
  
  devMastSetPortMode(0,DevMastRsMode_232);

  // echo data
  
  for (;;)
    {
      len = sizeof buf;
      status = cyg_io_read(ioHandle, buf, &len);

      // diag_printf("%d[%d]\n",portNum,len);

      // semi-blocking read shouldn't return with len of 0,
      // but let's check just the same

      if (len == 0)
        continue;
      
      totalBytes += len;

      if (status != ENOERR && status != -EAGAIN)
        {
          diag_printf("ERROR, cyg_io_read returned %d: %s\n",status,strerror(status));
          done();
        }

      status = cyg_io_write(ioHandle, buf, &len);

      if (status != ENOERR)
        {
          diag_printf("ERROR, cyg_io_write returned %d: %s\n",status,strerror(status));
          done();
        }
    }
}


static void bgTask(cyg_addrword_t data)
{
  cyg_thread_delay(300);  // wait a while...
  diag_printf("Extension background task running\n");
  while (1)
    {
      cyg_thread_delay(100);  // 1 second delay
      diag_printf("Bytes transfered: %d\n",totalBytes);
    }
}
