#include <pkgconf/hal.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_cache.h>
#include CYGHWR_MEMORY_LAYOUT_H
#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>
#include "flash.h"

// According to Intel datasheet upper 8 bits of command should
// be all 0s or all 1s

#define FLASH_Read_ID      0x0090
#define FLASH_Read_Status  0x0070
#define FLASH_Clear_Status 0x0050
#define FLASH_Program      0x0040
#define FLASH_Block_Erase  0x0020
#define FLASH_Suspend      0x00B0
#define FLASH_Resume       0x00D0
#define FLASH_Confirm      0x00D0
#define FLASH_Reset        0x00FF
#define FLASH_Config_Setup 0x0060
#define FLASH_Lock_Block   0x0001
#define FLASH_Unlock_Block 0x00D0

#define FLASH_Status_Ready 0x0080

#define FLASH_Intel_code   0x89

#ifndef GO_FAST
# define GO_FAST
# define GO_SLOW
#endif

extern volatile unsigned char EpldFlashWriteEnable;

//      Save manufacturer and device ID into two short ints.

static volatile unsigned short *ROM = (volatile unsigned short *)0x5000000;

static unsigned long FlashId;

int flash_query(unsigned short *data)
{
  EpldFlashWriteEnable = 1;
  ROM[0] = FLASH_Read_ID;
  data[0] = ROM[0];  // Manufacturer code
  data[1] = ROM[1];  // Device identifier
  FlashId = (ROM[0]<<16)|ROM[1];
  diag_printf("FlashId=%08x\n",FlashId);
  ROM[0] = FLASH_Reset;
  EpldFlashWriteEnable = 0;
  return 1;
}

static void lock_sector(volatile unsigned short *block)
{
  if (FlashId != 0x008988C5 && FlashId != 0x002088BB)
    return;
  diag_printf("lock %x\n",block);
  ROM[0] = FLASH_Config_Setup;
  *block = FLASH_Lock_Block;
}


static void unlock_sector(volatile unsigned short *block)
{
  if (FlashId != 0x008988C5 && FlashId != 0x002088BB)
    return;
  diag_printf("unlock %x\n",block);
  ROM[0] = FLASH_Config_Setup;
  *block = FLASH_Unlock_Block;
}

static void delayLoop(int i)
{
  while (i--)
    {
      asm(" nop");
      asm(" nop");
      asm(" nop");
    }
}

// macro to force an uncached read from external SDRAM in order to
// force the flash chip select inactive.  Numonyx flash requires the
// chip be deselected before reading status each time.  Otherwise, if
// we're running from cache, the status isn't read correctly.

#define flush() (*((volatile unsigned long*)0x4000000))

#define FlashStatus() (flush(),*ROM)

//      Erase the sector pointed to by "block"
  
int flash_erase_block(volatile unsigned short *block)
{
  externC cyg_ucount32 cyg_scheduler_read_lock(void);
  unsigned long stat;
  int timeout;
  int locked = cyg_scheduler_read_lock();
  GO_SLOW
  diag_printf("flash_erase_block ROM=%08x, block=%08x locked=%d\n", ROM, block, locked);
  EpldFlashWriteEnable = 1;
  unlock_sector(block);
  *ROM = FLASH_Clear_Status;
  *ROM = FLASH_Block_Erase;
  *block = FLASH_Confirm;
  timeout = 10000;
  while ((FlashStatus() & FLASH_Status_Ready) && timeout--)
    ;
  timeout = 100;
  while (!((stat=FlashStatus()) & FLASH_Status_Ready) && timeout--)
    {
      if (locked)
        delayLoop(50000);
      else
        cyg_thread_delay(1);
    }
  //diag_printf("flash erase time=%d0ms\n",100-timeout);
  lock_sector(block);
  *ROM = FLASH_Reset;
  EpldFlashWriteEnable = 0;
  GO_FAST 
  return stat == FLASH_Status_Ready;
}


int flash_program_buf(void *flashbase, void *rambase, int len)
{
  volatile unsigned short *last = (volatile unsigned short *)0;
  volatile unsigned short *addr = (volatile unsigned short *)flashbase;
  unsigned short *data = (unsigned short *)rambase;
  unsigned short stat = 0;
  unsigned timeout,i;

  EpldFlashWriteEnable = 1;

  *ROM = FLASH_Reset;

#if 0
  diag_printf("data\n");
  diag_dump_buf((void*)rambase,256);

  diag_printf("before\n");
  diag_dump_buf((void*)flashbase,256);
#endif

  GO_SLOW;
  diag_printf("flash_program(%08x,%08x,%d)\n", addr, data, len);

  *ROM = FLASH_Clear_Status;

  i = len;

  while (i > 0)
    {
      if ((((unsigned long)last)^((unsigned long)addr)) & ~0x1fff)
        {
          // crossed 8K boundary
          if (last)
            lock_sector(last);
          unlock_sector(addr);
          last = addr;
        }

      *addr = FLASH_Program;
      *addr = *data;

      timeout = 100;
      while ((!((stat=(flush(),*ROM)) & FLASH_Status_Ready)) && timeout--)
        delayLoop(10);

      if (timeout==0)
        {
          diag_printf("timeout: @0x%p stat=%04x\n",addr,stat);
          break;
        }

      if (stat & 0x007E)
        {
          diag_printf("bad status: @0x%p stat=%04x\n",addr,stat);
          break;
        }


      ++addr;
      ++data;
      i -= 2;
    }

  if (last)
    lock_sector(last);

  // verify programmed data

  *ROM = FLASH_Reset;

  addr = (volatile unsigned short *)flashbase;
  data = (unsigned short *)rambase;

  i = len;
  while (i)
    {
      if (*addr != *data)
	{
          diag_printf("program verify fail at %p: %04x %04x\n",addr,*addr,*data);
	  stat = 0x9910;
	  break;
	}
      ++addr;
      ++data;
      i -= 2;
    }

#if 0
  diag_printf("after\n");
  diag_dump_buf((void*)flashbase,256);
#endif

  GO_FAST;

  EpldFlashWriteEnable = 0;

  return stat == FLASH_Status_Ready;
}
