#include <cyg/hal/hal_cache.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/kernel/kapi.h>
#include <cyg/error/codes.h>
#include <cyg/infra/diag.h>
#include <cyg/io/io.h>
#include <cyg/io/serialio.h>
#include <cyg/io/file.h>
#include <net/etherio.h>
#include <stdlib.h>
#include <network.h>
#include <netinet/tcp.h>
#include <stdio.h>
#include <errno.h>

#include "../assert.h"
#include "../dm.h"
#include "../state.h"
#include "../ident.h"
#include "flash.h"

void setTcpKeepaliveTimeout(int timeout);
int getTcpKeepaliveTimeout(void);

extern int select(int, fd_set *, fd_set *, fd_set *, struct timeval *tv);
extern int cyg_select_with_abort(int, fd_set *, fd_set *, fd_set *, struct timeval *tv);
extern void cyg_select_abort(void);
extern int close(int);

static void portDtr(cyg_io_handle_t h, int dtrOn);
extern char *utoa(unsigned);
static int SocketIsClosed(int sock);


/* product ID info */
extern tIdent ident;

//Serial Hub 8 or 16  
extern int HAS_ONLY_232;

/* state for tcp accept() task */
static unsigned char AcceptStack[4096];
static cyg_handle_t AcceptHandle;
static cyg_thread AcceptThread;

/* global flag set by TCP/IP stack init routine */
extern int TcpStackOk;

/* declare the functions that run as tasks */
static cyg_thread_entry_t TcpRxTask, TcpTxTask, TcpAcceptTask;

/* atomic diag_printf operation -- only use in running tasks,
 * not in initialization code, DSR or ISR code.
 */
#define UseDiagPrintfMutex 0
#if UseDiagPrintfMutex
static cyg_mutex_t dpMutex;
#define diag_printf_m(fmt, args...) \
  do { \
       cyg_mutex_lock(&dpMutex); \
       diag_printf(fmt, ##args); \
       cyg_mutex_unlock(&dpMutex); \
     } while (0)
#else
#define diag_printf_m(fmt, args...) diag_printf(fmt, ##args)
#endif

/* diag printf macro called for debuggin r/w operations. */
#define DebugReadWrite 0
#if DebugReadWrite
#define rw_diag_printf_m(fmt, args...) diag_printf_m(fmt, ##args)
#else
#define rw_diag_printf_m(fmt, args...)	/* nothing */
#endif

typedef struct
{
    unsigned long configVersion;
    union
    {
	tPortConfig config;
	unsigned char padding[200];
    }u[DevMastMaxPorts];

    int tcpKeepaliveTimeout;
  
    /* add new stuff at bottom and assume they'll get defaulted to 0 */
}
tDmConfig;


/* typdef for config dat in flash to be stored in 8K flash sector*/
typedef struct
{
    unsigned long flags;
#define FlashConfigErased 0xffffffff
#define FlashConfigActive 0xfffffffe
#define FlashConfigInactive 0xfffffffc
    union
    {
	unsigned long rawdata[(0x2000 / (sizeof(unsigned long))) - 2];	// total size = 8K
	tDmConfig dmConfig;
    }u;
    unsigned long checksum;
}
tFlashConfig;

#define NumElements(a) ((sizeof (a))/(sizeof ((a)[0])))

static tFlashConfig *flashConfig0 = (tFlashConfig *) 0x5002000;
static tFlashConfig *flashConfig1 = (tFlashConfig *) 0x5004000;
static tFlashConfig *activeFlashConfig;

static unsigned long sum32(unsigned long *p, int count)
{
    unsigned long sum = 0;
    while (count--)
	sum += *p++;
    return sum;
}

int isActive(tFlashConfig * flashConfig)
{
    return flashConfig->flags == FlashConfigActive && flashConfig->u.dmConfig.configVersion == ConfigVersion && flashConfig->checksum == sum32(flashConfig->u.rawdata, NumElements(flashConfig->u.rawdata));
}

void serverWipeConfig(void)
{
    unsigned short id[2];
    flash_query(id);
    flash_erase_block((volatile unsigned short *) flashConfig0);
    flash_erase_block((volatile unsigned short *) flashConfig1);
}

static tDmConfig dmConfig;

int serverSaveConfig(void)
{
    int i, stat;
    unsigned short id[2];
    unsigned long u;
    tFlashConfig *next, *old;

    if (activeFlashConfig == flashConfig0)
    {
	next = flashConfig1;
	old = flashConfig0;
    }
    else
    {
	next = flashConfig0;
	old = flashConfig1;
    }

    diag_printf("Saving config to %08x\n", next);

    flash_query(id);
    if ((id[0] != 0x0089) && (id[0] != 0x0020))
    {
	diag_printf("Flash id wrong: %04x\n", id);
	return 3;
    }

    stat = flash_erase_block((volatile unsigned short *) next);

    memset(&dmConfig, 0, sizeof dmConfig);
  
    for (i = 0; i < ident.numPorts; ++i)
	dmConfig.u[i].config = serverState[i].config;
    dmConfig.configVersion = ConfigVersion;
    dmConfig.tcpKeepaliveTimeout = getTcpKeepaliveTimeout();

    flash_program_buf(&next->u.dmConfig, &dmConfig, sizeof dmConfig);
  
    if (memcmp(&next->u.dmConfig, &dmConfig, sizeof dmConfig))
    {
	diag_printf("dmConfig verify fail\n");
	return 1;
    }
  
    u = sum32(next->u.rawdata, NumElements(next->u.rawdata));
    flash_program_buf(&next->checksum, &u, sizeof next->checksum);
    if (next->checksum != u)
    {
	diag_printf("sum verify failed\n");
	return 2;
    }

    u = FlashConfigActive;
    flash_program_buf(&next->flags, &u, sizeof next->flags);

    u = FlashConfigInactive;
    flash_program_buf(&old->flags, &u, sizeof next->flags);

    activeFlashConfig = next;
    diag_printf("flash write done.\n");
    return 0;
}

/*
 *  Init function for the SocketServer application.
 */
void serverInit(void)
{
    int i;

    /* tell the world we're alive */
    diag_printf("socket server init\n");

    /* look for config info in flash */
    diag_printf("tPortConfig: %d/%d\n",sizeof (tPortConfig),sizeof (activeFlashConfig->u.dmConfig.u[0]));
    assert(sizeof (tPortConfig) <= sizeof (activeFlashConfig->u.dmConfig.u[0]));

    diag_printf("tDmConfig: %d/%d\n",sizeof (tDmConfig),sizeof (activeFlashConfig->u.rawdata));
    assert(sizeof (tDmConfig) <= sizeof (activeFlashConfig->u.rawdata));
  
    diag_printf("tFlashConfig: %d/%d\n",sizeof (tFlashConfig), 0x2000);
    assert(sizeof (tFlashConfig) <= 0x2000);
  
    if (isActive(flashConfig0))
	activeFlashConfig = flashConfig0;
    else if (isActive(flashConfig1))
	activeFlashConfig = flashConfig1;

    if (activeFlashConfig) 
        diag_printf("Flash active config found: %08x\n", activeFlashConfig);
    else
        diag_printf("Flash config area uninitialized (no active config found), using defaults... \n");
    diag_printf("\n");

     /* initialize mutex used by diag_printf_m() */
#if UseDiagPrintfMutex
    cyg_mutex_init(&dpMutex);
#endif

    if (activeFlashConfig)
    {
	setTcpKeepaliveTimeout(activeFlashConfig->u.dmConfig.tcpKeepaliveTimeout);
    }
    else
    {
	setTcpKeepaliveTimeout(60);
    }
  
    for (i = 0; i < ident.numPorts; ++i)
    {
	/* initialize server state */
	cyg_mutex_init(&serverState[i].serverMutex);
	cyg_semaphore_init(&serverState[i].socketSemaphore, 0);
	serverState[i].serverNumber = i;
	serverState[i].socket = -1;
	serverState[i].state = ServerIdle;
	serverState[i].configUpdated = 1;

	/*  
	 *  If activeFlashConfig is NULL, assume empty flash, start with default values.  Else,
	 *  use the stored values.
	 */  
	if (activeFlashConfig) {
	    serverState[i].config = activeFlashConfig->u.dmConfig.u[i].config;
	}
	else
	{
	    memset(&serverState[i].config,0,sizeof serverState[i].config);
	    serverState[i].config.baudRate = 9600;
	    serverState[i].config.interfaceMode = Mode232;
	    serverState[i].config.flowControl = FlowNone;
	    serverState[i].config.parity = ParityNone;
	    serverState[i].config.listenPort = 8000 + i;
	    serverState[i].config.enabled = 0;
	    serverState[i].config.wordLength = 8;
	    serverState[i].config.stopBits = 1;
	    serverState[i].config.listenEnable = 1;
	    serverState[i].config.idleTimeout = 300;
	}

	serverState[i].rxCount = 0;
	serverState[i].txCount = 0;
      
	/* serial port handle already looked up for us */
	serverState[i].serialHandle = devMastPort[i].handle;
	serverState[i].ownership = &devMastPort[i].ownership;

	/* Create TCP server tasks: one Rx (socket->serial) and one Tx (serial->socket) for each serial port */
	cyg_thread_create(9, TcpRxTask, (cyg_addrword_t) (serverState + i), "TcpRx", 
			  (void *) serverState[i].rxStack, sizeof serverState[i].rxStack, 
			  &serverState[i].rxHandle, &serverState[i].rxThread);
	cyg_thread_resume(serverState[i].rxHandle);

	cyg_thread_create(10, TcpTxTask, (cyg_addrword_t) (serverState + i), "TcpTx", 
			  (void *) serverState[i].txStack, sizeof serverState[i].txStack, 
			  &serverState[i].txHandle, &serverState[i].txThread);
	cyg_thread_resume(serverState[i].txHandle);
    }

    /* create task to accept TCP connections */
    cyg_thread_create(6, TcpAcceptTask, (cyg_addrword_t) 0, "Accept", (void *) AcceptStack, sizeof AcceptStack, &AcceptHandle, &AcceptThread);
    cyg_thread_resume(AcceptHandle);
}


static void initSerialPort(int i)
{
    cyg_uint32 len;
    Cyg_ErrNo status;
    cyg_serial_info_t serConfig;
    tServerState *s = serverState + i;

    if (s->config.enabled)    
	s->configUpdated = 0;

    len = sizeof serConfig;
    status = cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_INFO, &serConfig, &len);

    if (status != ENOERR)
    {
	diag_printf_m("initSerialPort(%d): cyg_io_get_config returned %d: %s\n", s->serverNumber, status, strerror(-status));
	return;
    }

    serConfig.baud = s->config.baudRate;

    serConfig.inter_char_timeout = 0;
    serConfig.read_timeout = 100;  // 100ms timeout if no data

    serConfig.flags = 0;
    serConfig.flags |= CYG_SERIAL_FLAGS_RD_NONBLOCK;
  
    switch (s->config.parity)
    {
    default:
    case ParityNone:
	serConfig.parity = CYGNUM_SERIAL_PARITY_NONE;
	break;
    case ParityEven:
	serConfig.parity = CYGNUM_SERIAL_PARITY_EVEN;
	break;
    case ParityOdd:
	serConfig.parity = CYGNUM_SERIAL_PARITY_ODD;
	break;
    case ParityMark:
	serConfig.parity = CYGNUM_SERIAL_PARITY_MARK;
	break;
    case ParitySpace:
	serConfig.parity = CYGNUM_SERIAL_PARITY_SPACE;
	break;
    }

    switch (s->config.flowControl)
    {
    default:
    case FlowNone:
	break;
    case FlowRTSCTS:
	serConfig.flags |= CYG_SERIAL_FLAGS_RTSCTS;
	break;
    case FlowXonXoff:
	serConfig.flags |= (CYG_SERIAL_FLAGS_XONXOFF_RXFLOW + CYG_SERIAL_FLAGS_XONXOFF_TXFLOW);
    case FlowHalfDuplex:
	serConfig.flags |= (CYG_SERIAL_FLAGS_RTS_TOGGLE + CYG_SERIAL_FLAGS_CTS_TXFLOW);
	break;
    }

    switch (s->config.wordLength)
    {
    case 5:
	serConfig.word_length = CYGNUM_SERIAL_WORD_LENGTH_5;
	break;
    case 6:
	serConfig.word_length = CYGNUM_SERIAL_WORD_LENGTH_6;
	break;
    case 7:
	serConfig.word_length = CYGNUM_SERIAL_WORD_LENGTH_7;
	break;
    default:
    case 8:
	serConfig.word_length = CYGNUM_SERIAL_WORD_LENGTH_8;
	break;
    }

    switch (s->config.stopBits)
    {
    case 2:
	serConfig.stop = CYGNUM_SERIAL_STOP_2;
	break;
    default:
    case 1:
	serConfig.stop = CYGNUM_SERIAL_STOP_1;
	break;
    }

    switch (s->config.interfaceMode)
    {
    default:
    case Mode232:
	if ((s->config.flowControl == FlowHalfDuplex) && !HAS_ONLY_232)
	    serConfig.interface_mode = DevMastRsMode_232I;	//        devMastSetPortMode(i,DevMastRsMode_232I);
	else
	    serConfig.interface_mode = DevMastRsMode_232;	//        devMastSetPortMode(i,DevMastRsMode_232);
	break;
    case Mode422:
	if( !HAS_ONLY_232 )
	    serConfig.interface_mode = DevMastRsMode_422;	//      devMastSetPortMode(i,DevMastRsMode_422); 
	break;
    case Mode485:
	if( !HAS_ONLY_232 )
        {
	    serConfig.interface_mode = DevMastRsMode_485;	//      devMastSetPortMode(i,DevMastRsMode_485);
	    serConfig.flags |= CYG_SERIAL_FLAGS_RTS_TOGGLE;
	}
	break;
    }


    len = sizeof serConfig;
    status = cyg_io_set_config(s->serialHandle, CYG_IO_SET_CONFIG_SERIAL_INFO, &serConfig, &len);
    if (status != ENOERR)
	diag_printf_m("initSerialPort(%d): cyg_io_set_config returned %d: %s\n", s->serverNumber, status, strerror(-status));

    if (s->config.dtrControl == DTRon || (s->state == ServerBusy && s->config.dtrControl == DTRconnected))
	portDtr(s->serialHandle, 1);
    else
	portDtr(s->serialHandle, 0);

    cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_INPUT_FLUSH, NULL, 0);
    cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_OUTPUT_FLUSH, NULL, 0);
}


static void connectionUp(tServerState * s, 
			 int sock,
			 unsigned long remIP, 
			 unsigned short remPort, 
			 unsigned long localIp, 
			 unsigned short localPort)
{
    int yes = 1;
    struct in_addr a;

    a.s_addr = remIP;
    s->locIP = localIp;
    s->locPort = localPort;
    s->remIP = remIP;
    s->remPort = remPort;
    if (s->socket >= 0)
	close(s->socket);
    s->socket = sock;

    s->state = ServerBusy;
    s->idleTimer = 0;
    if (getTcpKeepaliveTimeout() != 0)
	if(setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &yes, sizeof yes))
	    diag_printf_m("connectionUp: setsockopt SO_KEEPALIVE failed\n");

    if (s->config.dtrControl == DTRoff)
	portDtr(s->serialHandle, 0);
    else
	portDtr(s->serialHandle, 1);
    diag_printf("Serial port %d connected to IP:%s TCP Port:%d \n", s->serverNumber, inet_ntoa(a), remPort);
    cyg_semaphore_post(&s->socketSemaphore);	// wake up Rx task
}


/*
 * Tcp accept task: handles incoming tcp connections.
 */

static void TcpAcceptTask(cyg_addrword_t data)
{
    struct timeval tv;
    struct sockaddr_in client_addr, local;
    socklen_t client_len;
    tServerState *s;
    int ticksNow, ticksNext;
    int sock;
    int i, nsel;
    int maxfd;
    fd_set skset;
    int one = 1;

    tv.tv_sec = 1;
    tv.tv_usec = 0;

    client_len = sizeof client_addr;

    diag_printf_m("TcpAccept started\n");

    while (!TcpStackOk)
	cyg_thread_delay(100);

    ticksNext = cyg_current_time() + 100;

    while (1)
    {
	FD_ZERO(&skset);
	maxfd = 0;

	// update sockets
	for (s = serverState, i = 0; i < ident.numPorts; ++i, ++s)
	{
	    int semVal;
          
	    cyg_semaphore_peek(s->ownership, &semVal);

	    // close old one if we don't want it any more
	    if ((s->state == ServerIdle &&  s->socket >=0) && 
		(!s->config.enabled || 
		 !s->config.listenEnable || 
		 s->lastListenPort != s->config.listenPort || 
		 !semVal))
	    {
		// no longer supposed to listen on this one
		diag_printf("TcpAccept %d: closing socket on port %d\n", i, s->lastListenPort);
		close(s->socket);
		s->socket = -1;
	    }

	    // new listener socket if needed
	    if (s->config.enabled && 
		s->config.listenEnable && 
		s->socket < 0 && 
		s->state == ServerIdle && 
		semVal)
	    {
		s->lastListenPort = s->config.listenPort;
		diag_printf("TcpAccept %d: TCP Port %d ready for connection \n", i, s->lastListenPort);
		initSerialPort(s->serverNumber);

		sock = socket(AF_INET, SOCK_STREAM, 0);

		if (sock < 0)
		{
		    diag_printf_m("TcpAccept %d: socket() failed: %d\n", i, strerror(-sock));
		    cyg_thread_delay(20);
		    continue;
		}

		memset(&local, 0, sizeof(local));
		local.sin_family = AF_INET;
		local.sin_len = sizeof(local);
		local.sin_port = ntohs(s->lastListenPort);
		local.sin_addr.s_addr = INADDR_ANY;
		if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)))
		    diag_printf_m("TcpAccept %d: setsockopt REUSEADDR failed - %s \n", i, strerror(errno));
		else if (setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, &one, sizeof(one)))
		    diag_printf_m("TcpAccept %d: setsockopt REUSEPORT failed - %s \n", i, strerror(errno));
		else if (bind(sock, (struct sockaddr *) &local, sizeof(local)) < 0)
		    diag_printf_m("TcpAccept %d: bind() failed: %d\n", i, strerror(errno));
		else if (listen(sock, 0))
		    diag_printf_m("TcpAccept %d: listen() failed: %d\n", i, errno);
		else
		    s->socket = sock;
	    }

	    if (s->socket >= 0 && s->state == ServerIdle)
	    {
		FD_SET(s->socket, &skset);
		if (s->socket > maxfd)
		    maxfd = s->socket;
	    }
	}

	// wait for connections
	nsel = cyg_select_with_abort(maxfd + 1, &skset, 0, 0, &tv);
	ticksNow = cyg_current_time();
	if ((ticksNext - ticksNow) < 0)
	{
	    // one second
	    ticksNext += 100;
	    for (s = serverState, i = 0; i < ident.numPorts; ++i, ++s)
		s->idleTimer += 1;
	}

	if (nsel <= 0)
	    continue;

	for (s = serverState, i = 0; i < ident.numPorts; ++i, ++s)
	{
	    if (!(s->socket >= 0 && FD_ISSET(s->socket, &skset)))
		continue;
          
	    client_len = sizeof client_addr;
	    sock = accept(s->socket, (struct sockaddr *) &client_addr, &client_len);
	    close(s->socket);
	    s->socket = -1;
	    if (sock < 0)
	    {
		diag_printf_m("TcpAccept %d: accept failed errno=%d\n", s->serverNumber, errno);
		cyg_thread_delay(20);	// maybe let network catch up?
	    }
	    else
	    {
		cyg_mutex_lock(&s->serverMutex);
		if (!s->config.enabled || 
		    !s->config.listenEnable || 
		    s->state != ServerIdle || 
		    !cyg_semaphore_trywait(s->ownership))
		{
		    diag_printf("TcpAccept %d: rejected\n", i);
		    close(sock);
		    cyg_mutex_unlock(&s->serverMutex);
		    continue;
		}

		/*  Successfull TCP connection has occurred */
//		diag_printf("TcpAccept %d: connection from %08x:%d\n",  s->serverNumber, ntohl(client_addr.sin_addr.s_addr), ntohs(client_addr.sin_port));
		if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one)))
		    diag_printf("TcpAccept:  Cannot set NoDelay\n");

		// flush port
		initSerialPort(s->serverNumber);
		connectionUp(s,
			     sock,
			     ntohl(client_addr.sin_addr.s_addr),
			     ntohs(client_addr.sin_port),
			     ntohl(INADDR_ANY),
			     s->lastListenPort);
		cyg_mutex_unlock(&s->serverMutex);
	    }
	}
    }
}


static void portDtr(cyg_io_handle_t h, int dtrOn)
{
    cyg_uint32 state = 0;
    cyg_uint32 len = sizeof state;
    if (dtrOn)
	state |= CYG_SERIAL_MODEM_DTR;
    cyg_io_set_config(h, CYG_IO_SET_CONFIG_SERIAL_MODEM, &state, &len);
}


/*
 *   Tcp receive task: transfer data from socket to serial port
 *   If no TCP connection is open (s->state == ServerIdle) this task blocks waiting 
 *   for the semaphore from TcpTaskAccept.
 */
static void TcpRxTask(cyg_addrword_t data)
{
    int n;
    char rxDataBuf[1024];
    char *p;
    Cyg_ErrNo status;
    tServerState *s = (tServerState *) data;
    static int rxBytes;

    while (1)
    {
	while (s->state == ServerBusy)
	{
	    // transfer data from TCP to serial port
	    n = read(s->socket, rxDataBuf, sizeof rxDataBuf);
	    rxBytes += n;
	    rw_diag_printf_m("TcpRx%d: rd %d total %d \n", s->serverNumber, n, rxBytes);

	    if (s->state != ServerBusy)
		break;

	    if (n == 0)
	    {
		// Connection is gone
		diag_printf_m("TcpRx%d: connection closed\n", s->serverNumber);
		s->state = ServerClosing;
		rxBytes = 0;
	    }

	    else if (n < 0)
	    {
		// read() error
		diag_printf("TcpRx%d: read() error %d - %s\n", s->serverNumber, n, strerror(-n));
		s->state = ServerClosing;
	    }

	    else
	    {
		// write data to serial port
		s->idleTimer = 0;
		s->rxCount += n;
		p = rxDataBuf;
		while (n > 0)
		{
		    cyg_uint32 len = n;
		    status = cyg_io_write(s->serialHandle, p, &len);
		    rw_diag_printf_m("TcpRx%d: wr %d\n", s->serverNumber, len);
		    if (status == -EINTR)
			break;  // somebody aborted the read/write
		    if (status != ENOERR)
		    {
			diag_printf_m("TcpRx%d: error writing to serial port: %d: %s\n", s->serverNumber, status, strerror(-status));
			cyg_thread_delay(20);
		    }
		    n -= len;
		    p += len;
		}
	    }
	}
//	diag_printf_m("TcpRx%d: waiting on semaphore\n", s->serverNumber);
	cyg_semaphore_wait(&s->socketSemaphore);
//	diag_printf_m("TcpRx%d: thread awoken \n", s->serverNumber);
    }
}


/*
 *   Tcp transmit task: transfers data from serial port to Tcp connection.
 *   If no TCP connection is open (s->state == ServerIdle) this task loops in the
 *   while(1) continuously, with the 50ms delay at the start.
 */
#define TxBufSize (8*1024)

static void TcpTxTask(cyg_addrword_t data)
{
    tServerState *s = (tServerState *) data;
    unsigned char *txDataBuf;
    unsigned txBufCount = 0;

    unsigned char *p;
    cyg_uint32 len;
    Cyg_ErrNo status;
    int semVal;
    int n;
    unsigned loops = 0;
    static int txBytes;

    // stagger thread start-up just for fun
    cyg_thread_delay(11 * s->serverNumber);
    diag_printf_m("TcpTx%d started\n", s->serverNumber);

    txDataBuf = malloc(TxBufSize);
  
    if (!txDataBuf)
    {
	diag_printf("ACK! malloc failed: %s %d\n",__FILE__,__LINE__);
	return;
    }
  
    while (1)
    {
	// delay for a bit -- so that we don't swamp the net with tiny packets.

	++loops;
	cyg_thread_delay(5);

	if (s->config.enabled && s->state == ServerIdle)
	{
	    /* check to see if we should initiate a connection */
	    int initiate = 0;

	    cyg_semaphore_peek(s->ownership, &semVal);

	    if (semVal)
	    {
		if (s->configUpdated)
		    initSerialPort(s->serverNumber);

		if (s->config.connectAlways)
		    initiate = 1;

		if (s->config.connectOnData)
		{
		    cyg_serial_buf_info_t serbufinfo;
		    len = sizeof serbufinfo;
		    status = cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_BUFFER_INFO, &serbufinfo, &len);
		    //diag_printf("TcpTx%d rx_count=%d\n",s->serverNumber, serbufinfo.rx_count);
		    if (status != ENOERR)
			diag_printf("TcpTx%d serial get buffer info error %d - %s\n", s->serverNumber, status, strerror(-status));
		    else if (serbufinfo.rx_count)
			initiate = 1;
		}

		if (s->config.connectOnCD || s->config.connectOnDSR)
		{
		    int msr;
		    len = sizeof msr;
		    status = cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_MODEM, &msr, &len);
		    if (status != ENOERR)
			diag_printf("TcpTx%d serial get modem status error %d\n", s->serverNumber, status);
		    else if (((msr & CYG_SERIAL_MODEM_CD) && s->config.connectOnCD) || ((msr & CYG_SERIAL_MODEM_DSR) && s->config.connectOnDSR))
			initiate = 1;
		}
	    }
	    if (initiate)
	    {
		int sock;
		struct sockaddr_in dest;
              
		memset(&dest, 0, sizeof(dest));
		dest.sin_family = AF_INET;
		dest.sin_port = htons(s->config.connectToPort);
		dest.sin_addr.s_addr = htonl(s->config.connectToIP);
              
		sock = socket(PF_INET, SOCK_STREAM, 0);
              
		if (sock < 0)
                {
		    diag_printf("TcpTx%d: socket() failed: %d\n", s->serverNumber, strerror(errno));
		    cyg_thread_delay(100);
		    continue;
                }
              
		if (connect(sock,(struct sockaddr *)&dest,sizeof dest) < 0)
                {
		    diag_printf("TcpTx%d: connect() failed: %s \n", s->serverNumber, strerror(errno));
		    close(sock);
		    cyg_thread_delay(100);
		    continue;
                }
              
		diag_printf("TcpTx%d: initiating\n",s->serverNumber);
              
		cyg_mutex_lock(&s->serverMutex);
		if (s->state == ServerIdle && cyg_semaphore_trywait(s->ownership))
		    connectionUp(s,sock, s->config.connectToIP,s->config.connectToPort,0,0);
		else
                {
		    diag_printf("TcpTx%d: acquire failure -- closing %d\n", s->serverNumber, sock);
		    close(sock);
                }
		cyg_mutex_unlock(&s->serverMutex);
	    }
	}

	if (s->state == ServerBusy && s->config.disconnOnIdle && (s->idleTimer > s->config.idleTimeout))
	{
	    diag_printf("TcpTx%d: idle timeout\n", s->serverNumber);
	    s->state = ServerClosing;
	}

	if (s->state == ServerBusy && (s->config.disconnOnNoCD || s->config.disconnOnNoDSR))
	{
	    int msr;

	    len = sizeof msr;
	    status = cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_MODEM, &msr, &len);
	    if (status != ENOERR)
		diag_printf("TcpTx%d serial get modem status error %d - %s \n", s->serverNumber, status, strerror(-status));
	    else if (((~msr & CYG_SERIAL_MODEM_CD) && s->config.disconnOnNoCD) || 
		     ((~msr & CYG_SERIAL_MODEM_DSR) && s->config.disconnOnNoDSR))
		s->state = ServerClosing;
	}

	if (s->state == ServerBusy && SocketIsClosed(s->socket))
        {
	    diag_printf("TcpTx%d: socket closed by remote host\n",s->serverNumber);
	    s->state = ServerClosing;
        }
      
	if (s->state == ServerBusy && !s->config.enabled)
	    s->state = ServerClosing;

	if (s->state == ServerClosing)
	{
	    cyg_mutex_lock(&s->serverMutex);

	    diag_printf("TcpTx%d: Closing socket \n", s->serverNumber);
	    close(s->socket);
	    s->socket = -1;
	    txBytes = 0;
	    txBufCount = 0;
	    s->locIP = 0;
	    s->locPort = 0;
	    s->remIP = 0;
	    s->remPort = 0;
	    s->idleTimer = 0;
	    s->state = ServerIdle;
	    cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_INPUT_FLUSH, NULL, 0);
	    cyg_io_get_config(s->serialHandle, CYG_IO_GET_CONFIG_SERIAL_OUTPUT_FLUSH, NULL, 0);
	    if (s->config.dtrControl == DTRon)
		portDtr(s->serialHandle, 1);
	    else
		portDtr(s->serialHandle, 0);
	    cyg_mutex_unlock(&s->serverMutex);
	    cyg_semaphore_post(s->ownership);
	    cyg_select_abort();	// wake up the accept task
	}

	/* Read data from serial port and shove into TCP connection */
	while (s->state == ServerBusy && s->config.enabled)
	{
	    if (s->configUpdated)
		initSerialPort(s->serverNumber);
          
	    len = TxBufSize - txBufCount;
	    status = cyg_io_read(s->serialHandle, txDataBuf+txBufCount, &len);
          
	    if (status != ENOERR && 
		status != -EAGAIN && 
		status != -EEOL && 
		status != -ETIMEDOUT &&
		status != -EINTERCHARTIMEOUT)
	    {
		diag_printf_m("TcpTx%d: cyg_io_read of %08x returned %d: %s\n", s->serverNumber, s->serialHandle, status, strerror(-status));
		cyg_thread_delay(10);
	    }
          
	    if (status == -EAGAIN && len == 0)
		break;

	    if (len)
		rw_diag_printf_m("TcpTx%d: rd %d (%d)\n", s->serverNumber, len, status);
          
	    txBufCount += len;
	    s->txCount += len;
	    s->idleTimer = 0;
          
	    // If we're waiting for IC Timeout and timed
	    // out waiting for one, then leave data in buffer.
	    if (status == -ETIMEDOUT) 
		break;
          
	    p = txDataBuf;
	    while (txBufCount)
            {
		n = write(s->socket,p,len);
		if (n < 0)
                {
		    diag_printf_m("TcpTx%d: pWrite() returned %d: %s\n", s->serverNumber, n, strerror(-n));
		    if (s->state == ServerBusy)
			s->state = ServerClosing;
		    break;
                }
		txBytes += n;
		rw_diag_printf_m("TcpTx%d: wr %d total %d\n", s->serverNumber, n, txBytes);
		p += n;
		txBufCount -= n;
            }
	}
    }
}


//===================
#include <sys/protosw.h>
static int tcpKeepaliveTimeout;


#if !defined(CYGPKG_NET_FREEBSD_STACK)
void setTcpKeepaliveTimeout(int timeout)
{
    extern int tcp_keepidle;
    extern int tcp_keepintvl;
    extern int tcp_maxpersistidle;
    extern int tcp_maxidle;
    extern int tcptv_keep_init;

    tcpKeepaliveTimeout = timeout;
    diag_printf("TCP Keepalive timeout %d seconds\n", timeout);

    if (timeout == 0)
	timeout = 60;

    if (timeout < 4)
	tcpKeepaliveTimeout = timeout = 4;

    timeout *= PR_SLOWHZ;

    tcp_maxidle = timeout / 2;
    tcp_keepidle = timeout / 2;
    tcp_keepintvl = timeout / 4;
    tcptv_keep_init = timeout / 2;
    tcp_maxpersistidle = timeout;
}
#else
void setTcpKeepaliveTimeout(int timeout)
{
    tcpKeepaliveTimeout = timeout;
    diag_printf("STUB: setTcpKeepaliveTimeout()\n");
    diag_printf("TCP Keepalive timeout %d seconds\n", timeout);
}
#endif

int getTcpKeepaliveTimeout(void)
{
    return tcpKeepaliveTimeout;
}

static int SocketIsClosed(int sock)
{
#if !defined(SO_CANTRCVMORE)
    return 0;
#else
    int ret, i, l;

    ret = getsockopt(sock, SOL_SOCKET, SO_CANTRCVMORE, &i, &l);
    if (ret == -1) {
	diag_printf("Error in SocketIsClosed getsocketopt - sock: %d  error: %d  - %s \n", sock, ret, strerror(-ret));
	return 0;
    }

    return i;
#endif
}
